local f = CreateFrame('frame', nil, WorldMapPlayerLower)
f:SetAllPoints()

-- This is all for drawing a line on the map in zones that we're not currently in
-- We could just draw our line outside of the scrollframe and rotate it, and to be honest that may even be a better approach
local Transforms = {}
for _, transformID in ipairs(GetWorldMapTransforms()) do
	local mapID, displayMapID, _, _, bottom, top, right, left, y, x = GetWorldMapTransformInfo(transformID)
	tinsert(Transforms, {
		mapID = mapID, displayMapID = displayMapID,
		left = left, top = top,
		right = right, bottom = bottom,
		x = x, y = y,
	})
end

local function TransformCoordinatePair(mapID, x, y)
	for transformID, transformData in pairs(Transforms) do
		if transformData.mapID == mapID and transformData.left > x and x > transformData.right and transformData.top > y and y > transformData.bottom then
			mapID = transformData.displayMapID
			-- todo: make sure this is right
			x, y = x + transformData.x, y + transformData.y
			break
		end
	end
	return mapID, x, y
end

local function TransformCoordinates(mapID, left, top, right, bottom)
	local isTransformed = false
	for transformID, transformData in pairs(Transforms) do
		if transformData.mapID == mapID and transformData.left > left and right > transformData.right and transformData.top > top and bottom > transformData.bottom then
			mapID = transformData.displayMapID
			left, top = left + transformData.x, top + transformData.y
			right, bottom = right + transformData.x, bottom + transformData.y
			isTransformed = true
			break
		end
	end
	return mapID, left, top, right, bottom, isTransformed
end

local function GetMapSize() -- Return dimensions and offset of current map
	local _, left, top, right, bottom = GetCurrentMapZone()
	local floorNum, dright, dbottom, dleft, dtop = GetCurrentMapDungeonLevel()
	if DungeonUsesTerrainMap() then floorNum = floorNum - 1 end
	if floorNum > 0 then left, top, right, bottom = dleft, dtop, dright, dbottom end
	
	if left and left ~= right then
		local width, height = left - right, top - bottom
		local isTransformed = false
		local mapID, _, displayMapID = GetAreaMapInfo(GetCurrentMapAreaID())
		if displayMapID ~= -1 then
			mapID, left, top, right, bottom, isTransformed = TransformCoordinates(mapID, left, top, right, bottom)
		end
		
		return left, top, right, bottom, width, height, mapID, isTransformed --format('%d.%d', GetCurrentMapAreaID(), floorNum)
	end
end

local function GetIntersect(px, py, a, sx, sy, ex, ey)
	a = (a + PI / 2) % (PI * 2)
	local dx, dy = -math.cos(a), math.sin(a)
	local d = dx * (sy - ey) + dy * (ex - sx)
	if d ~= 0 and dx ~= 0 then
		local s = (dx * (sy - py) - dy * (sx - px)) / d
		if s >= 0 and s <= 1 then
			local r = (sx + (ex - sx) * s - px) / dx
			if r >= 0 then
				return sx + (ex - sx) * s, sy + (ey - sy) * s, r, s
			end
		end
	end
end

local LineFrame = CreateFrame('frame', nil, WorldMapButton)
LineFrame:SetAllPoints()
LineFrame:SetFrameLevel(150)

-- These frames are essentially just placeholders to anchor our line to
local StartPoint = CreateFrame('frame', nil, LineFrame) StartPoint:SetSize(1, 1)
local EndPoint = CreateFrame('frame', nil, LineFrame) EndPoint:SetSize(1, 1)

local Line = LineFrame:CreateLine(nil, 'OVERLAY')
Line:Hide()
Line:SetTexture('interface/buttons/white8x8')
Line:SetGradientAlpha('HORIZONTAL', 0, 0, 0, 0.5, 1, 0, 0, 0.8)
Line:SetThickness(2)
Line:SetStartPoint('CENTER', StartPoint, 0, 0)
Line:SetEndPoint('CENTER', EndPoint, 0, 0)

local WorldMapUpdated, PlayerFacing = false, 0
LineFrame:SetScript('OnUpdate', function(self, elapsed)
	local angle = GetPlayerFacing()
	if WorldMapUpdated or GetUnitSpeed('player') > 0 or angle ~= PlayerFacing then
		WorldMapUpdated = false
		PlayerFacing = angle
		Line:Hide()
		
		if UnitOnTaxi('player') then -- hide line while in flight
			return
		end
		
		local mx, my = GetPlayerMapPosition('player')
		if not mx then return end
		local onMap = true
		if mx == 0 and my == 0 then
			onMap = false
		end
		local py, px, _, pMapID = UnitPosition('player')
		if not px then return end -- we somehow do not have coordinates
		local left, top, right, bottom, width, height, mapMapID, isTransformed = GetMapSize()
		if not width or width == 0 then return end -- map has no size?
		if isTransformed then
			pMapID, px, py = TransformCoordinatePair(pMapID, px, py)
		end
		local playerMapAreaID = GetPlayerMapAreaID('player')
		local sameInstanceish = false
		if playerMapAreaID and mapMapID ~= pMapID then
			local mapID, _, displayMapID = GetAreaMapInfo(GetCurrentMapAreaID())
			mapID = displayMapID == -1 and mapID or displayMapID
			
			local mapID2, _, displayMapID2 = GetAreaMapInfo(playerMapAreaID)
			mapID2 = displayMapID2 == -1 and mapID2 or displayMapID2
		
			sameInstanceish = mapID == mapID2
		end
		if mapMapID == pMapID or onMap or sameInstanceish then -- same instance
			-- this could probably be simplified, but there's probably no point
		
			-- top left to top right
			local topX, topY, topRi, topSi = GetIntersect(px, py, angle, left, top, right, top)
			-- bottom left to bottom right
			local bottomX, bottomY, bottomRi, bottomSi = GetIntersect(px, py, angle, left, bottom, right, bottom)
			-- top left to bottom left
			local leftX, leftY, leftRi, leftSi = GetIntersect(px, py, angle, left, top, left, bottom)
			-- top right to bottom right
			local rightX, rightY, rightRi, rightSi = GetIntersect(px, py, angle, right, top, right, bottom)
			
			local mx1, my1, mr1, ms1
			local mx2, my2, mr2, ms2
			local m1Side, m2Side -- top, bottom, left, right
			
			if topX then
				mx1, my1, mr1, ms1 = topX, topY, topRi, topSi
				m1Side = 'top'
			end
			if bottomX then
				if not mx1 then
					mx1, my1, mr1, ms1 = bottomX, bottomY, bottomRi, bottomSi
					m1Side = 'bottom'
				else
					mx2, my2, mr2, ms2 = bottomX, bottomY, bottomRi, bottomSi
					m2Side = 'bottom'
				end
			end
			if leftX then
				if not mx1 then
					mx1, my1, mr1, ms1 = leftX, leftY, leftRi, leftSi
					m1Side = 'left'
				else
					mx2, my2, mr2, ms2 = leftX, leftY, leftRi, leftSi
					m2Side = 'left'
				end
			end
			if rightX then
				if not mx1 then
					mx1, my1, mr1, ms1 = rightX, rightY, rightRi, rightSi
					m1Side = 'right'
				else
					mx2, my2, mr2, ms2 = rightX, rightY, rightRi, rightSi
					m2Side = 'right'
				end
			end
			
			if m1Side and m2Side then -- we have 2 points
				StartPoint:ClearAllPoints()
				EndPoint:ClearAllPoints()
				if mr1 < mr2 then -- m1 is closer, so use that as starting point
					if m1Side == 'top' then
						StartPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 1002 * ms1, 0)
					elseif m1Side == 'bottom' then
						StartPoint:SetPoint('CENTER', WorldMapButton, 'BOTTOMLEFT', 1002 * ms1, 0)
					elseif m1Side == 'left' then
						StartPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 0, -668 * ms1)
					elseif m1Side == 'right' then
						StartPoint:SetPoint('CENTER', WorldMapButton, 'TOPRIGHT', 0, -668 * ms1)
					end

					if m2Side == 'top' then
						EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 1002 * ms2, 0)
					elseif m2Side == 'bottom' then
						EndPoint:SetPoint('CENTER', WorldMapButton, 'BOTTOMLEFT', 1002 * ms2, 0)
					elseif m2Side == 'left' then
						EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 0, -668 * ms2)
					elseif m2Side == 'right' then
						EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPRIGHT', 0, -668 * ms2)
					end
				else -- m2 is closer
					if m2Side == 'top' then
						StartPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 1002 * ms2, 0)
					elseif m2Side == 'bottom' then
						StartPoint:SetPoint('CENTER', WorldMapButton, 'BOTTOMLEFT', 1002 * ms2, 0)
					elseif m2Side == 'left' then
						StartPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 0, -668 * ms2)
					elseif m2Side == 'right' then
						StartPoint:SetPoint('CENTER', WorldMapButton, 'TOPRIGHT', 0, -668 * ms2)
					end
					
					if m1Side == 'top' then
						EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 1002 * ms1, 0)
					elseif m1Side == 'bottom' then
						EndPoint:SetPoint('CENTER', WorldMapButton, 'BOTTOMLEFT', 1002 * ms1, 0)
					elseif m1Side == 'left' then
						EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 0, -668 * ms1)
					elseif m1Side == 'right' then
						EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPRIGHT', 0, -668 * ms1)
					end
				end
				Line:Show()
			elseif m1Side and onMap then
				StartPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 1002 * mx, -668 * my)
				if m1Side == 'top' then
					EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 1002 * ms1, 0)
				elseif m1Side == 'bottom' then
					EndPoint:SetPoint('CENTER', WorldMapButton, 'BOTTOMLEFT', 1002 * ms1, 0)
				elseif m1Side == 'left' then
					EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPLEFT', 0, -668 * ms1)
				elseif m1Side == 'right' then
					EndPoint:SetPoint('CENTER', WorldMapButton, 'TOPRIGHT', 0, -668 * ms1)
				end
				Line:Show()
			end
		end
	end
end)

LineFrame:SetScript('OnEvent', function(self, event, ...)
	WorldMapUpdated = true
end) LineFrame:RegisterEvent('WORLD_MAP_UPDATE')